{-# OPTIONS -fglasgow-exts #-}

{-| Position information for syntax. Crucial for giving good error messages.
-}

module Syntax.Position
  ( -- * Positions
    Posn(..)
  , startPos
  , unknownPos
  , movePos

    -- * Ranges
  , Range(..)
  , noRange
  , HasRange(..)
  , fuseRange
  ) where

import Data.Generics

-- | Represents a point in the input (file, line, col) or
--   an unknown position
data Posn   = Pn    { srcFile   :: FilePath
		    , posLine   :: !Int
		    , posCol    :: !Int
		    }
	    | NoPos
    deriving (Typeable, Data, Eq)

-- | A range is a pair of positions. The rEnd position is
--   not included in the range.
data Range = Range { rStart, rEnd :: Posn }
    deriving (Typeable, Data, Eq)

-- | Ranges between two unknown positions
noRange :: Range
noRange = Range unknownPos unknownPos

instance Show Posn where
    show (Pn "" l c)	= show l ++ "," ++ show c
    show (Pn f l c)	= f ++ ":" ++ show l ++ "," ++ show c
    show NoPos		= "<No position>"

instance Show Range where
    show (Range NoPos NoPos) = ""
    show (Range s e) = file ++ start ++ "-" ++ end
	where
	    f	= srcFile s
	    sl	= posLine s
	    el	= posLine e
	    sc	= posCol s
	    ec	= posCol e
	    file
		| null f    = ""
		| otherwise = f ++ ":"
	    start = show sl ++ "," ++ show sc
	    end
		| sl == el  = show ec
		| otherwise = show el ++ "," ++ show ec
			
-- | The first position in a file
startPos :: FilePath -> Posn
startPos f = Pn { srcFile = f, posLine = 1, posCol = 1 }

unknownPos :: Posn
unknownPos = NoPos

-- | Advance the position by one character.
--   A tab character ('\t') will move the position to the next
--   tabstop (tab size is 8). A newline character ('\n') moves
--   the position to the first character in the next line. Any
--   other character moves the position to the next column.
--   Does not traverse module boundaries.
movePos :: Posn -> Char -> Posn
movePos (Pn f l c) '\t'	= Pn f l (div (c + 7) 8 * 8 + 1)
movePos (Pn f l c) '\n'	= Pn f (l + 1) 1
movePos (Pn f l c) _	= Pn f l (c + 1)
movePos NoPos _		= NoPos

-- | Things that have a range are instances of this class. In particular
--   concrete and internal terms.
class HasRange t where
    getRange :: t -> Range

instance HasRange Range where
    getRange = id

-- | Finds the least interval that covers its arguments. The left argument
--   is assumed to be to the left of the right argument.
--   We cannot get rid of this assumption
--   since the file names are not ordered and hence Posn is not ordered
fuseRange :: (HasRange t, HasRange u) => t -> u -> Range
fuseRange x y = Range (rStart $ getRange x) (rEnd $ getRange y)

