{-# OPTIONS -fglasgow-exts #-}

module Plugins.FOL.Gandalf (gandalfProof, Mangling(..)) where

import Data.Char
import Data.List
import Text.ParserCombinators.Parsec
import Text.PrettyPrint

import Utils.Monads

{--------------------------------------------------------------------------
    Pretty proof
 --------------------------------------------------------------------------}

data Mangling = SantaMangling | AgdaLightMangling

gandalfProof :: Mangling -> String -> String
gandalfProof m output = unlines $ concatMap disp $ proof
    where
	d = case m of
		AgdaLightMangling   -> agdaLightDemangler
		SantaMangling	    -> santaDemangler
	disp s = [showClause d $ parseClause s]
	proof =
	    takeWhile ("END OF PROOF" /=)
	    $ drop 1
	    $ dropWhile ("START OF PROOF" /=)
	    $ lines output


{--------------------------------------------------------------------------
    Lexer
 --------------------------------------------------------------------------}

data Token  = TkInt Int
	    | TkJustification String
	    | TkId String
	    | TkSym Char
	    | TkContradiction
    deriving (Eq, Show)

isSym c = elem c "().,-|?"
isId c	= isAlphaNum c || elem c "_^'"

contradiction = "contradiction"

lexer :: String -> [Token]
lexer s
    | contradiction `isPrefixOf` s
		    = TkContradiction : lexer (drop (length contradiction) s)
lexer ('[':s)	    = case span (/=']') s of
			(s1,_:s2)   -> TkJustification s1 : lexer s2
			_	    -> error "impossible: Plugin.FOL.Gandalf.lexer"
lexer ('$':s)	= case span isId s of
			(s1,s2) -> TkId ('c':'$':s1) : lexer s2
lexer (c:s)
    | isSpace c = lexer s
    | isSym c	= TkSym c : lexer s
    | isDigit c = case span isDigit s of
			(s1,s2)	-> TkInt (read $ c:s1) : lexer s2
    | isId c	= case span isId s of
			(s1,s2)	-> TkId (c:s1) : lexer s2
lexer []	= []
lexer s		= error $ "lex error in Gandalf proof on input: " ++ s

{--------------------------------------------------------------------------
    Parser
 --------------------------------------------------------------------------}

type P = GenParser Token ()

data Clause	= Clause Int String [Literal]
data Literal	= Neg Atom
		| Pos Atom
		| Unknown
		| Contradiction
data Atom	= Predicate String [Term]
data Term	= Var String
		| Fun String [Term]

showClause :: Demangler -> Clause -> String
showClause d = show . pretty d

parseClause s =
    case parse clauseP "" (lexer s) of
	Right c	    -> c
	Left err    -> error $ show err

thisP t = tokenPrim show (\p _ _ -> p) f
    where
	f t'	| t' == t   = Just ()
	f _		    = Nothing

symP c = thisP (TkSym c)

clauseP :: P Clause
clauseP =
    do	TkInt n		    <- anyToken
	TkJustification j   <- anyToken
	ls <-	fmap (const [Contradiction]) (thisP TkContradiction)
		<|> fmap (const [Unknown]) (symP '?')
		<|> do	ls <- chainr1 (fmap (:[]) litP) orP
			symP '.'
			return ls
	return $ Clause n j ls
    where
	orP = const (++) <$> symP '|'

litP :: P Literal
litP =
    do	symP '-'
	Neg <$> atomP
    <|>	Pos <$> atomP

atomP :: P Atom
atomP =
    do	TkId x <- anyToken
	ts  <- argsP
	return $ Predicate x ts

argsP = option [] $
	    between (symP '(') (symP ')')
		    (fmap (const []) (thisP (TkInt 0))
			<|>
		     sepBy termP (symP ',')
		    )

termP :: P Term
termP =
    do	TkId x <- anyToken
	if isVar x
	    then return $ Var x
	    else do ts <- argsP
		    return $ Fun x ts
    where
	isVar (c:s) = isUpper c
	isVar _	    = False

{------------------------------------------------------------------------
    Demangling
 ------------------------------------------------------------------------}

data Demangler =
	Demangler   { isApp	    :: String -> Bool
		    , isEqual	    :: String -> Bool
		    , isPair	    :: String -> Bool
		    , isProj	    :: String -> Bool
		    , isFst	    :: String -> Bool
		    , isSnd	    :: String -> Bool
		    , isInfixFun    :: String -> Bool
		    , isInfixPred   :: String -> Bool
		    , demangleVar   :: String -> String
		    , demangleFun   :: String -> String
		    , demanglePred  :: String -> String
		    }

santaDemangler :: Demangler
santaDemangler =
    Demangler	{ isApp		= (==) "f_^a^p_"
		, isEqual	= (==) "equal"
		, isPair	= (==) "pair_" . demangle
		, isProj	= (==) "proj_" . demangle
		, isFst		= (==) "fst_" . demangle
		, isSnd		= (==) "snd_" . demangle
		, isInfixPred	= isOperator . demangle
		, isInfixFun	= isOperator . demangle
		, demangleVar	= id
		, demangleFun	= demangle
		, demanglePred	= demangle
		}
    where
	demangle = quote . demngl . up . tail

	demngl "_TY_"	    = "::"
	demngl "_BT_"	    = "_|_"
	demngl "_THIS_IS_AN_EMPTY_CLAUSE_"  = "FALSE"
	demngl "_AP_"	    = "$app"
	demngl x
	    | Just (y,i) <- isPrimed x  = y ++ show i
	    | last x == '_'		= deOp $ init $ tail x
	    | otherwise			= x

	isPrimed :: String ->  Maybe (String, Int)
	isPrimed = ip . reverse
	    where
		ip ('_':x) =
		    case span isDigit x of
			(n@(_:_),'p':'_':y) ->
			     Just (demngl $ reverse $ '_':y, read $ reverse n)
			_ -> Nothing
		ip _ = Nothing

	deOp s
	    | Just (c,s')   <- match s ops  = c : deOp s'
	    | '_':a:b:s'    <- s	    = fromHex a b : deOp s'
	    | c:s'	    <- s	    = c : deOp s'
	    | otherwise			    = []
	    where
		match s [] = Nothing
		match s ((c,op):ops)
		    | isPrefixOf op s	= Just (c, drop (length op) s)
		    | otherwise		= match s ops

		fromHex a b = toEnum (16 * hex a + hex b)
		    where
			hex c	| '0' <= c && c <= '9'	= read [c]
				| 'A' <= c && c <= 'F'	= fromEnum c - fromEnum 'A' + 10
				| otherwise		= error $ "fromHex: " ++ show s

	ops =
	    [ ('_', "__")
	    , ('"', "_qu")
	    , ('!', "_ex")
	    , ('?', "_qu")
	    , ('#', "_ha")
	    , ('@', "_at")
	    , ('$', "_S_")
	    , ('%', "_pc")
	    , ('\'', "_pr")
	    , ('&', "_an")
	    , ('|', "_or")
	    , ('/', "_fw")
	    , ('\\', "_bk")
	    , ('*', "_ti")
	    , ('+', "_pl")
	    , ('-', "_mi")
	    , ('>', "_gt")
	    , ('=', "_eq")
	    , ('<', "_lt")
	    , ('^', "_up")
	    , ('~', "_tw")
	    , (':', "_cn")
	    , (';', "_sc")
	    , (',', "_cm")
	    , ('.', "_o_")
	    ]

agdaLightDemangler :: Demangler
agdaLightDemangler =
    Demangler	{ isApp		= (==) "capp_"
		, isEqual	= (==) "equal"
		, isPair	= (==) "cpair_"
		, isProj	= (==) "cproj_"
		, isFst		= (==) "cfst_"
		, isSnd		= (==) "csnd_"
		, isInfixPred	= isPrefixOf op__
		, isInfixFun	= isPrefixOf ('c':op__)
		, demangleVar	= id
		, demangleFun	= reop . up . tail
		, demanglePred	= reop . up
		}
    where
	op__ = "op__"

	reop s
	    | op__ `isPrefixOf` s   = parseOp $ drop (length op__) s
	    | otherwise		    = s
	    where

		trans _ ""		    = ""
		trans (op,name) s@(c:s')
		    | name `isPrefixOf` s   = op : trans (op,name) (drop (length name) s)
		    | otherwise		    = c : trans (op,name) s'

		parseOp s = foldr trans s ops

		ops =
		    [ ('!'  , "bang")
		    , ('#'  , "hash")
		    , ('$'  , "dollar")
		    , ('%'  , "percent")
		    , ('&'  , "amp")
		    , ('*'  , "times")
		    , ('+'  , "plus")
		    , ('.'  , "dot")
		    , ('/'  , "slash")
		    , ('<'  , "less")
		    , ('='  , "eq")
		    , ('>'  , "greater")
		    , ('?'  , "question")
		    , ('@'  , "at")
		    , ('\\' , "backslash")
		    , ('^'  , "hat")
		    , ('|'  , "bar")
		    , ('-'  , "minus")
		    , ('~'  , "twiddle")
		    , (':'  , "colon")
		    ]

up ('^':c:s)	= toUpper c : up s
up (c:s)	= c : up s
up []		= []

quote x
    | isValid x	    = x
    | isOperator x  = x
    | otherwise	    = "\"" ++ x ++ "\""
    where
	isValid (c:s) = (isAlpha c || c == '_') && all ok s
	    where
		ok '_'	= True
		ok '\'' = True
		ok c	= isAlphaNum c
	isValid _ = error "impossible: Plugins.FOL.Gandalf.isValid []"

isOperator (c:s) = elem c opchars && all (`elem` ('\'':opchars)) s
    where
	opchars = "!#$%&*+./<=>?@\\^|-~:"
isOperator _ = error "impossible: Plugins.FOL.Gandalf.isOperator []"

{--------------------------------------------------------------------------
    App view
 --------------------------------------------------------------------------}

data AppTerm
	= AppVar String [Term]
	| AppFun String [Term]

appView d (Var x) = AppVar x []
appView d (Fun a [s,t])
    | isApp d a   = app (appView d s) t
    where
	app (AppVar x ts) t = AppVar x (ts ++ [t])
	app (AppFun x ts) t = AppFun x (ts ++ [t])
appView d (Fun f ts)	    = AppFun f ts

{--------------------------------------------------------------------------
    Pretty
 --------------------------------------------------------------------------}

class Pretty a where
    pretty	:: Demangler -> a -> Doc
    prettyPrec	:: Demangler -> Int -> a -> Doc

    pretty d	    = prettyPrec d 0
    prettyPrec d _  = pretty d

instance Pretty Int where
    pretty _ = text . show

instance Pretty Clause where
    pretty d (Clause n j ls) =
	case partition isNeg ls of
	    (ns@(_:_), ps) ->
		pretty d n <+>
		    sep	[ fsep [ list "/\\" $ map pos ns
			       , text " ==> "
			       , orMe ps
			       ]
			, nest 20 $ prJ j
			]

	    (_, _)	->
		pretty d n <+>
		    sep	[ orMe ls
			, nest 20 $ prJ j
			]
	where
	    orMe [] = text "_|_"
	    orMe xs = list "\\/" xs

	    list op xs = fsep $ intersperse (text op) $ map (pretty d) xs

	    prJ ""  = empty
	    prJ j   = text "[" <> text j <> text "]"

	    pos (Neg x) = Pos x
	    pos _	= error $ "impossible: Plugins.FOL.Gandalf.pos"

	    isNeg (Neg _) = True
	    isNeg _	  = False

instance Pretty Literal where
    pretty d (Neg (Predicate equal [s,t]))
	| isEqual d equal   = prettyOp d 2 (text "!=") s t 0
    pretty d (Pos (Predicate equal [s,t]))
	| isEqual d equal   = prettyOp d 2 (text "==") s t 0
    pretty d (Neg a)	    = text "-" <> prettyPrec d 9 a
    pretty d (Pos a)	    = pretty d a
    pretty d Unknown	    = text "?"
    pretty d Contradiction  = text "_|_"

instance Pretty Atom where
    prettyPrec d p (Predicate x [s,t])
	| isInfixPred d x = prettyOp d 6 (prettyPred d x) s t p
    prettyPrec d p (Predicate x ts) = pApp d p (prettyPred d x) ts

prettyOp d p' op x y p =
    mparens (p>=p') $
	sep [ prettyPrec d p' x <+> op
	    , prettyPrec d p' y
	    ]

instance Pretty Term where
    prettyPrec d p t =
	case appView d t of
	    AppFun pair [s,t]
		| isPair d pair	    -> parens $ pretty d s <> comma <> pretty d t
	    AppFun proj (t:pr:ts)
		| isProj d proj	    -> pApp d p (pProj d t pr) ts
	    AppFun x [s,t]
		| isInfixFun d x    -> prettyOp d 6 (prettyFun d x) s t p
	    AppFun x ts		    -> pApp d p (prettyFun d x) ts
	    AppVar x ts		    -> pApp d p (prettyVar d x) ts

mparens b = if b then parens else id

pProj d t (Fun p []) = prettyPrec d 10 t <> text s
    where
	s   | isFst d p	= ".1"
	    | isSnd d p	= ".2"
	    | otherwise	= p
pProj _ _ _ = error "impossible: Plugins.FOL.Gandalf.pProj"

pApp d p x [] = x
pApp d p x ts = mparens (p>9) $ x <+> argList ts
    where
	argList ts = fsep $ map (prettyPrec d 10) ts

prettyVar d	= text . demangleVar d
prettyFun d	= text . demangleFun d
prettyPred d	= text . demanglePred d

