{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Monad where

import Control.Monad	    ( liftM )
import Control.Monad.State
import Control.Monad.Error
import Control.Monad.Reader
import Data.Set (Set)
import Data.Generics
import System.IO.Unsafe

import Interface.Options    ( Options(..) )

import Syntax.Internal
import Syntax.Values
import Syntax.Position	    ( Range )
import Syntax.Sorts
import Syntax.ConcreteToInternal (ScopeEnv, emptyScopeEnv)

import TypeChecker.Context
import {-# SOURCE #-} TypeChecker.Rewriting

import Utils.Error
import Utils.Map
import Utils.Monads
import Utils.Names

import {-# SOURCE #-} Plugins	-- REMOVE_IF_HADDOCK

{- REMOVE_IF_HADDOCK
-- | This datatype is not really here. It is just an illusion to fool
--   Haddock into accepting recursive modules. The real datatype is
--   'Plugins.SomePlugin'.
data SomePlugin
REMOVE_IF_HADDOCK -}

-- The type checking monad ------------------------------------------------

-- | The type checking monad
--type TC = ReaderT Environment (StateT TCState (ErrorT ErrorMsg IO))
data TC a = TC { unTC :: ReaderT Environment
			    (StateT TCState
			    (ErrorT ErrorMsg IO))
			    a
		  }

data TCClosure a = Unevaluated Environment TCState (TC a)
		 | Evaluated a

buildTCClosure :: TC a -> TC (TCClosure a)
buildTCClosure m =
    do	env <- ask
	st  <- get
	return $ Unevaluated env st m

-- | Running a closure doesn't affect the current state.
evalTCClosure :: TCClosure a -> TC a
evalTCClosure (Unevaluated env st m) =
    do	st' <- get
	put st
	x <- local (const env) m
	put st'
	return x
evalTCClosure (Evaluated x) = return x

unsafeEvalTCClosure :: TCClosure a -> a
unsafeEvalTCClosure (Unevaluated env st m) =
    unsafePerformIO $
	do  r <- runTC $ do put st
			    local (const env) m
	    either (error . show) return r
unsafeEvalTCClosure (Evaluated x) = x

instance Monad TC where
    return  = TC . return
    m >>= k = TC $ unTC m >>= unTC . k
    fail    = TC . fail

instance MonadReader Environment TC where
    ask	    = TC ask
    local f = TC . local f . unTC

instance MonadState TCState TC where
    get	    = TC get
    put	    = TC . put

instance MonadError ErrorMsg TC where
    throwError	    = TC . throwError
    catchError m h  = TC $ catchError (unTC m) (unTC . h)

instance MonadIO TC where
    liftIO  = TC . liftIO

-- | The environment used in type checking.
data Environment = Environment
    { envContext	:: Context
	-- ^ The current context. Maps names to types.
    , envErrorMsg	:: ErrorMsg
	-- ^ The current error message. Contains information about
	--   the current activity and position in the source code,
	--   so that should an error occur, we can give a nice message.
    , envOptions	:: Options
    , envLocalRewriteRules  :: RewriteRules
	-- ^ Command line options.
    , envPolymorphicLevels	:: Bool
    , envAllowUnsolvedMetas	:: Bool
    , envFirstOrderUnification	:: Bool
    , envIsLocalDefinition	:: Bool
    , envScopeEnv		:: ScopeEnv
	-- ^ In a local definition we don't require meta variables to
	--   be solved.
    }

emptyEnv :: Environment
emptyEnv = Environment [] noMsg undefined empty True False False False emptyScopeEnv

data TCState = TCState
    { tcUsedNames		:: UsedNames
    , tcPlugins			:: Map String SomePlugin
    , tcDefinitions		:: Definitions
	-- ^ The definitions in scope at the moment.
    , tcDatatypes		:: Datatypes
	-- ^ All datatypes in scope.
    , tcSignatures		:: Signatures
	-- ^ All signatures
    , tcGlobalRewriteRules	:: RewriteRules
    , tcMetaVars		:: MetaVars
	-- ^ Keeping track of meta variables for error messages
    , tcLevelConstraints	:: [SortConstraint]
    , tcPostponedConstraints	:: [(MetaRef, [(Environment,TC ())])]
    , tcStatistics		:: Statistics
    }

initState :: TCState
initState = TCState 
	    { tcUsedNames		= noNamesUsed
	    , tcPlugins			= empty
	    , tcDefinitions		= empty
	    , tcDatatypes		= empty
	    , tcSignatures		= empty
	    , tcGlobalRewriteRules	= empty
	    , tcMetaVars		= []
	    , tcLevelConstraints	= []
	    , tcPostponedConstraints	= []
	    , tcStatistics		= empty
	    }

clearState :: TC ()
clearState = modify $ const initState

runTC :: TC a -> IO (Either ErrorMsg a)
runTC tc =
    runErrorT
	$ flip evalStateT initState
	$ flip runReaderT emptyEnv
	$ unTC tc

useFirstOrderUnification :: TC a -> TC a
useFirstOrderUnification = local $ \e -> e { envFirstOrderUnification = True }

usePatternUnification :: TC a -> TC a
usePatternUnification = local $ \e -> e { envFirstOrderUnification = False }

isFirstOrderUnification :: TC Bool
isFirstOrderUnification = envFirstOrderUnification <$> ask

isLocalDefinition :: TC Bool
isLocalDefinition = envIsLocalDefinition <$> ask

localDefinition :: TC a -> TC a
localDefinition = local $ \e -> e { envIsLocalDefinition = True }

noPolymorphicLevels :: TC a -> TC a
noPolymorphicLevels = local (\e -> e { envPolymorphicLevels = False })

polymorphicLevels :: TC Bool
polymorphicLevels = envPolymorphicLevels <$> ask

getLevelConstraints :: TC [SortConstraint]
getLevelConstraints = tcLevelConstraints <$> get

addLevelConstraints :: [SortConstraint] -> TC ()
addLevelConstraints cs =
    modify $ \s -> s { tcLevelConstraints = cs ++ tcLevelConstraints s }

getScopeEnv :: TC ScopeEnv
getScopeEnv = envScopeEnv <$> ask

withScopeEnv :: ScopeEnv -> TC a -> TC a
withScopeEnv e = local $ \s -> s { envScopeEnv = e }

type Definitions = Map Name PolymorphicDefValue

-- | Map from datatype name to constructors and their definitions.
type Datatypes = Map Name [(Name,DefValue)]
type Signatures = Map Name SigInfo

data SigInfo = SigInfo { sigFields  :: [(Name, FieldSource, DefValue)]
		       , sigParents :: Set Name
		       }
    deriving (Show)

data FieldSource = InheritedFrom Name
		 | NewField
    deriving (Show, Eq)

-- | The set of all userdefined sorts. The type of a sort is stored
--   in the 'Definitions' component of the 'Environment'.
type Sorts = Set Name

-- Meta variables ---------------------------------------------------------

data MetaVarInfo = MVI	Range	-- where
			Term	-- which function is it applied to
			Int	-- which argument is it
			Value	-- what's the instantiation
type MetaVars = [MetaVarInfo]

-- Statistics -------------------------------------------------------------

type Statistics = Map Stat Integer
type Stat = String

-- Names ------------------------------------------------------------------

setUsedNames :: UsedNames -> TC ()
setUsedNames used = modify (setNames used)

getUsedNames :: TC UsedNames
getUsedNames = getNames <$> get

instance HasNames TCState where
    getNames	= tcUsedNames
    setNames x	= \st -> st { tcUsedNames = x }

